#ifndef ATOMICSPARC_H
#define ATOMICSPARC_H

/* returns the orginal value of var (before the add is applied) */
static inline int fetchAndAdd(volatile int & var, int addend)             
{                                                                       
  int orig, added;
  bool result;

  do {
    orig = var;
    added = orig + addend;
    result = compareAndSwap(var, orig, added);
  } while (!result);
  
  return orig;                                                   
} 

/* CAS macro
   semantics:  
   If (a == b) then a = c
   c = OLD value of a
*/    
static inline bool compareAndSwap(volatile int & a, int b, int c)
{
  __asm__ __volatile__("mov %1,%%l2\n\t"
                       "mov %2,%%l1\n\t"
                       "mov %0,%%l3\n\t"   
                       "cas [%%l2],%%l1,%%l3\n\t" 
                       "mov %%l3, %0"
                       :"=r"(c) 
                       : "r"(&a), "r"(b),"r"(c)	
                       :"l1","l2","l3","memory"
                       );
  if(c == b) return true;
  else return false;
}

#endif // ATOMICSPARC_H
