#ifndef ATMOICX86_H
#define ATOMICX86_H

#include "unistd.h"

static inline unsigned long long int fetchAndAdd(volatile unsigned long long int & var, long long int addend)
{
#if defined(__x86_64__) || defined(__x86_64) || defined(__amd64__) || defined(__amd64)
  __asm__ __volatile__("lock\n\t"
                       "xaddq %0,%1"
                       : "+r"(addend), "+m"(var)
                       : 
                       : "memory");
  return static_cast<unsigned long long int> (addend);
#elif defined(__tune_i686__)
  // can't do a 64-bit fetch&incr on a 32-bit machine
  // but we can do compare and swap!
  unsigned long long int result;
  unsigned long long int orig;
  do {
    orig = var;
    result = var + addend;
  } while (!compareAndSwap(var, orig, result));
  return orig;
#else
#error "Architecture not supported"
#endif
}

static inline long long int fetchAndAdd(volatile long long int & var, long long int addend)
{
#if defined(__x86_64__) || defined(__x86_64) || defined(__amd64__) || defined(__amd64)
  __asm__ __volatile__("lock\n\t"
                       "xaddq %0,%1"
                       : "+r"(addend), "+m"(var)
                       : 
                       : "memory");
  return addend;
#elif defined(__tune_i686__)
  // can't do a 64-bit fetch&incr on a 32-bit machine
  // but we can do compare and swap!
  long long int result;
  long long int orig;
  do {
    orig = var;
    result = var + addend;
  } while (!compareAndSwap(var, orig, result));
  return orig;
#else
#error "Architecture not supported"
#endif
}

static inline int fetchAndAdd(volatile int & var, int addend)
{
  /**
   *  xadd a b;  
   *     a => b, 
   *     b => a + b
   */
  __asm__ __volatile__("lock\n\t"
                       "xaddl %0,%1"
                       : "+r"(addend), "+m"(var)
                       : 
                       : "memory"); 
  return addend;
} 

static inline bool compareAndSwap(volatile int & a, int b, int c)
{
  bool ret;
  __asm__ __volatile__
    (
     "lock\n\t"
     "cmpxchg %1,%2\n\t"
     "sete %0\n\t"
     :
     "=q"(ret)
     :
     "q"(c),
     "m"(a),
     "a"(b)
     :
     "memory", "cc"
     );
  return ret;
}

static inline int64_t tbb_cmpswp8 (volatile void *ptr, int64_t value, int64_t comparand )
{
    const int32_t comparand_lo = (int32_t)comparand;
    const int32_t comparand_hi = *(int32_t*)((intptr_t)&comparand+sizeof(int32_t));
    int64_t result;
    // EBX register saved for compliancy with position-independent code (PIC) rules on IA32
    __asm__ __volatile__ (
             "pushl %%ebx\n\t"
             "movl  (%%ecx),%%ebx\n\t"
             "movl  4(%%ecx),%%ecx\n\t"
             "lock\ncmpxchg8b (%2)\n\t"
             "popl  %%ebx"
                    : "=A"(result), "=m"(*(int64_t *)ptr)
                    : "S"(ptr),
                      "a"(comparand_lo),
                      "d"(comparand_hi),
                      "c"(&value)
                    : "memory", "esp");
    return result;
}

static inline bool compareAndSwap(volatile long long int & a, long long int b, long long int c)
{
  int result = tbb_cmpswp8(&a,c,b);
  return result == b;
}

static inline bool compareAndSwap(volatile unsigned long long int & a, unsigned long long int b, unsigned long long int c)
{
  bool ret;
  __asm__ __volatile__
    (
     "lock\n\t"
     "cmpxchg8b %5\n\t"
     "sete %0\n\t"
     :
     "=q" (ret)
     :
     "a" (((unsigned long*)&b)[0]),
     "d" (((unsigned long*)&b)[1]),
     "b" (((unsigned long*)&c )[0]),
     "c" (((unsigned long*)&c )[1]),
     "m" (a)
     :
     "memory", "cc"
     );
  return ret;
}

#endif // ATOMICX86_H
