
#ifndef COARSE_LOCK_Q_H
#define COARSE_LOCK_Q_H

#include "AbstractQueue.h"
#include "Vector.h"

#include <pthread.h>

#ifdef TBB_SUPPORTED
#include <tbb/spin_mutex.h>
#endif

#include <deque>

#include <iostream>

class AbstractResult;

extern bool g_queue_empty;


/* NOTE: the scoped spin lock is provided for your use, but you do not
   have to utilize it */


#ifdef TBB_SUPPORTED

typedef tbb::spin_mutex taskman_mutex_t;

#elif defined(__sparc) || defined(sparc) || defined(__sparc__)

class taskman_sparc_spinlock{
private:
  unsigned char* m_lock;
  
public:
  taskman_sparc_spinlock(){
    m_lock = new unsigned char;
    *m_lock = 0;
  }
  ~taskman_sparc_spinlock(){
    delete m_lock;
  }
  
  class scoped_lock{
  public:
    
    unsigned char** l;
    
    scoped_lock( ) 
      : l(NULL) {}
    scoped_lock( taskman_sparc_spinlock& tolock ) 
      : l(&tolock.m_lock) 
    {
      __asm__ __volatile__ ("\n1:\n\t"
                            "ldstub [%0], %%g2\n\t"
                            "orcc %%g2, 0x0, %%g0\n\t"
                            "bne,a 2f\n\t"
                            " ldub [%0], %%g2\n\t"
                            ".subsection   2\n"
                            "2:\n\t"
                            "orcc  %%g2, 0x0, %%g0\n\t"
                            "bne,a  2b\n\t"
                            ".previous\n"
                            : /* no outputs */
                            : "r"(*l)
                            : "g2", "memory", "cc");
      
    }
    
    ~scoped_lock(){
      if(l) { __asm__ __volatile__("stb %%g0, [%0]" : : "r" (*l): "memory"); }
    }
    
    bool try_acquire( taskman_sparc_spinlock& tolock ){
      bool ret = false;
#ifdef TASKMAN_DEBUG
      assert( m_lock == NULL );
#endif
      l = &tolock.m_lock;
      if( l ){
	unsigned int result;
	__asm__ __volatile__("ldstub [%1], %0"
			     : "=r" (result)
			     : "r" (*l)
			     : "memory");
	ret = (result == 0);
      }
      if(!ret) { l = NULL; }
      return ret;
    }
    
  };
};

typedef taskman_sparc_spinlock taskman_mutex_t;

#else
class taskman_mutex_wrapper{

private:
  pthread_mutex_t m_lock;

public:  
  taskman_mutex_wrapper(){
    pthread_mutex_init( &m_lock, NULL );
  }
  ~taskman_mutex_wrapper(){
    pthread_mutex_destroy( &m_lock );
  }

  class scoped_lock{
  public:

    pthread_mutex_t* l;
    
    scoped_lock( ) : l(NULL)
    {}

    scoped_lock( taskman_mutex_wrapper& tolock ):l(&tolock.m_lock)
    {
      pthread_mutex_lock( l );
    }

    ~scoped_lock(){ 
      if( l ){ pthread_mutex_unlock( l ); }
    }

    bool try_acquire( taskman_mutex_wrapper& tolock ){

      bool ret = false;
      assert( l == NULL );
      l = &tolock.m_lock;
      if( l ){
        ret = (pthread_mutex_trylock(l) == 0);
      }
      if(!ret){ l = NULL; }

      return ret;
    }

  };

  //friend class scoped_lock;

};

typedef taskman_mutex_wrapper taskman_mutex_t;

#endif

template< typename T >
class CoarseLockQueue : public AbstractQueue<T>
{
public:
  CoarseLockQueue() : AbstractQueue<T> () {
  }

  virtual const char* getName(){ 
    return "CoarseGrainedLockQueue";
  }

  virtual void enqueue(T element){
  }

  virtual bool dequeue(T& element){ 
    return false;
  }


  virtual bool peekFront(T& element){
    return false;
  }

  virtual bool peekBack(T& element){ 
    return false;
  } 



protected:
  virtual bool steal(T& element){ 
    return false;
  }

protected:

public:
  static void initStatics( void ){ 

  };

};

#endif /* COARSE_LOCK_Q_H */
