
#include <tbb/task.h>
#include <tbb/task_scheduler_init.h>
#include <tbb/tick_count.h>
#include <iostream>

using namespace tbb;

long SerialFib( long n ) {
  if(n<2)
    return n;
  else
    return SerialFib(n-1) + SerialFib(n-2);
}

class FibTask: public task {
public:
  long* const sum;
  const long n;

  FibTask( long _n, long* _sum ) :
    sum(_sum), n(_n)
  {}

  task* execute(){
    if(n < 16) {
      *sum = SerialFib(n);
    } else {
      long x, y;
      FibTask& a = *new( allocate_child() ) FibTask(n-1, &x);
      FibTask& b = *new( allocate_child() ) FibTask(n-2, &y);
      set_ref_count(3);
      spawn(b);
      spawn_and_wait_for_all(a);
      *sum = x+y;
    }
    return NULL;
  }


};

int main( int argc, char* argv[] )
{
  long sum;
  long n;
  long num_procs;
  tick_count start_time, end_time;

  assert(argc == 3);
  n = atoi(argv[1]);
  num_procs = atoi(argv[2]);

  task_scheduler_init init(num_procs);

  start_time = tick_count::now();

  FibTask& a = *new( task::allocate_root() ) FibTask(n, &sum);
  task::spawn_root_and_wait(a);

  end_time = tick_count::now();

  std::cout << "sum = " << sum << std::endl;
  std::cout << "time = " << (end_time - start_time).seconds() << "s" << std::endl;

  return 0;
}
