
#include <iostream>
#include <unistd.h>
#include <ctime>
#include <stdio.h>

#include "Distributions.h"
#include "Timers.h"
#include <TaskMan.h>
//#include "../tbb/include/tbb/atomic.h"

/* test types */
#define LOOP 1
#define TAIL_TASK 2
#define HEAD_TASK 3

using namespace std;

//tbb::atomic< hrtime_t > total_work_time;
hrtime_t total_work_time = 0;
int total_tasks = 0;
int g_max_depth;
int g_node_breadth;

int loopTest(Distribution* dist)
{
  float work_time_fp;
  hrtime_t work_time_delta;
  hrtime_t running_time;
  hrtime_t start_time, end_time;

  start_time = gethrtime();

  work_time_fp = dist->getSample();
  work_time_delta = get_hrtime_delta(work_time_fp);

  do {
    end_time = gethrtime();
    running_time =  end_time - start_time;
  } while ( running_time < work_time_delta );

  end_time = gethrtime();
  fetchAndAdd(total_work_time, end_time - start_time);
  fetchAndAdd(total_tasks, 1);
  //  total_work_time.fetch_and_add(end_time - start_time);

  return 1;
}

int taskHeadRecursionTest(Distribution* dist, int depth)
{
  float work_time_fp;
  hrtime_t work_time_delta;
  hrtime_t running_time;
  hrtime_t start_time, end_time;

  if(depth < g_max_depth){
    result<int>* results = new result<int>[g_node_breadth];
    int dummy;
    for(int i=0;i<g_node_breadth;i++)
      results[i] = task(taskHeadRecursionTest, dist, depth+1);
    for(int i=0;i<g_node_breadth;i++)
      dummy = *results[i];
  }

  start_time = gethrtime();

  work_time_fp = dist->getSample();
  work_time_delta = get_hrtime_delta(work_time_fp);

  do {
    end_time = gethrtime();
    running_time =  end_time - start_time;
  } while ( running_time < work_time_delta );

  end_time = gethrtime();
  fetchAndAdd(total_work_time, end_time - start_time);
  fetchAndAdd(total_tasks, 1);

  return 1;
}

int taskTailRecursionTest(Distribution* dist, int depth)
{
  float work_time_fp;
  hrtime_t work_time_delta;
  hrtime_t running_time;
  hrtime_t start_time, end_time;
  
  start_time = gethrtime();

  work_time_fp = dist->getSample();
  work_time_delta = get_hrtime_delta(work_time_fp);

  do {
    end_time = gethrtime();
    running_time =  end_time - start_time;
  } while ( running_time < work_time_delta );

  end_time = gethrtime();
  fetchAndAdd(total_work_time, end_time - start_time);
  fetchAndAdd(total_tasks, 1);

  if(depth < g_max_depth){
    result<int>* results = new result<int>[g_node_breadth];
    int dummy;
    for(int i=0;i<g_node_breadth;i++)
      results[i] = task(taskTailRecursionTest, dist, depth+1);
    for(int i=0;i<g_node_breadth;i++)
      dummy = *results[i];
  }

  return 1;
}

void printHelp(char* bin_name)
{
  cerr << "Usage: " << bin_name << " [OPTIONS]" << endl;
    cerr << "  Required: " << endl;
    cerr << "       -t <Test type> <param 1 [ param 2 ] > " << endl;
    cerr << "           Loop:         l <# tasks> " << endl;
    cerr << "           TaskTail:     t <tree depth> <node breadth> " << endl;
    cerr << "           TaskHead:     h <tree depth> <node breadth> " << endl;
    cerr << "       -d <Distribution> <param 1 [ param 2 ] > " << endl;
    cerr << "           Uniform:      u <max>" << endl;
    cerr << "           Exponential:  e <lambda>" << endl;
    cerr << "           Pareto:       p <alpha> <k>" << endl;
    cerr << "       -n <# threads> " << endl << endl;
    cerr << "  Optional:" << endl;
    cerr << "       -s <seed value | 0 for random>" << endl;
    cerr << "       -q <queue type>" << endl;
    cerr << "           FineGrainedLock: l {default} " << endl;
    cerr << "           Xaction:         x " << endl;
    cerr << "           Lock-Free:       w " << endl;
}

int main(int argc, char* argv[])
{
  int opt;
  Distribution* dist = NULL;
  int num_tasks, num_threads;
  int test_type = LOOP;
  int seed;

  num_tasks = 10000; // default to something
  num_threads = 3;
  seed = time(NULL);

  set_queue_type(COARSE_LOCK_QUEUE);

  if( argc < 2 ){
    printHelp(argv[0]);
    exit(EXIT_FAILURE);
  }
  while((opt = getopt(argc, argv, "d:n:t:s:q:h")) != -1){
    switch(opt){
    case 'd':
      if(optarg[0] == 'e'){
        float lambda;
        lambda = atof( argv[optind] );
        optind++;
        dist = new ExponentialDistribution(lambda);
      } else if(optarg[0] == 'p'){
        float alpha, k;
        alpha = atof( argv[optind] );
        assert(optind <= argc-1);
        optind++;
        k = atof( argv[optind] );
        optind++;
        dist = new ParetoDistribution(alpha, k);
      } else if(optarg[0] == 'u'){
        int max;
        max = atoi( argv[optind] );
        optind++;
        dist = new UniformDistribution(max);
      } else {
        cerr << "Unkown distribution type:" << optarg << endl;
        exit(EXIT_FAILURE);
      }
      break;
    case 'h':
      printHelp(argv[0]);
      exit(EXIT_SUCCESS);
    case 'n':
      num_threads = atoi(optarg);
      break;
    case 's':
      seed = atoi(optarg);
      break;
    case 't':
      if(optarg[0] == 'l'){
        test_type = LOOP;
        num_tasks = atoi( argv[optind] );
        optind++;
      } else if(optarg[0] == 't'){
        test_type = TAIL_TASK;
        g_max_depth = atoi( argv[optind] );
        optind++;
        g_node_breadth = atoi( argv[optind] );
        optind++;
      } else if(optarg[0] == 'h'){
        test_type = HEAD_TASK;
        g_max_depth = atoi( argv[optind] );
        optind++;
        g_node_breadth = atoi( argv[optind] );
        optind++;
      } else {
        cerr << "Unkown test type:" << optarg << endl;
        exit(EXIT_FAILURE);
      }
      break;
    case 'q':
      if( optarg[0] == 'l'){
        set_queue_type(COARSE_LOCK_QUEUE);
        optind++;
        break;
      } else if( optarg[0] == 'x' ){
        set_queue_type(XACTION_QUEUE);
        optind++;
      } else if( optarg[0] == 'w' ){
        set_queue_type(LOCK_FREE_QUEUE);
        optind++;
      } else {
        cerr << "Unkown queue type: " << optarg << endl;
        exit(EXIT_FAILURE);
      }
      break;
    default:
      cerr << "Unkown Argument: " << opt << "\n";
      exit(EXIT_FAILURE);
    };
  }
  if( dist == NULL ){ printHelp(argv[0]); }

  srand(seed);

  cout << "Running with parameters: " << endl;
  cout << "   Test: " << test_type << endl;
  cout << "   " << dist->getName() << " distribution - mean: " << dist->getMean() << " std deviation: " << dist->getStdDeviation() << endl;
  cout << "   Queue Type: " << get_queue_type_name() << endl;
  cout << "   Seed value: " << seed << endl;
  cout << "   Number of tasks:   " << num_tasks << endl;
  cout << "   Number of threads: " << num_threads << endl;

  taskman_init(num_threads);

  //int total;

  hrtime_t start_time, end_time;
  result<int>* results = new result<int>[num_tasks];
  result<int> res;
  int dummy;
  start_time = gethrtime();

  if(test_type == LOOP){
    for( int i = 0; i< num_tasks; i++ ){
      results[i] = task( loopTest, dist );
    }
    taskman_sync();
  } else if(test_type == TAIL_TASK){
    res = task( taskTailRecursionTest, dist, 0 );
    dummy = *res;
  } else if(test_type == HEAD_TASK){
    res = task( taskHeadRecursionTest, dist, 0 );
    dummy = *res;
  } else {
    cerr << "Unkown test type" << endl;
    exit(EXIT_FAILURE);
  }

  end_time = gethrtime();
  std::cout << "Total runtime (s):       " << get_time_delta_s(end_time, start_time) << std::endl;
  std::cout << "Total working time (s):  " << get_time_delta_s(total_work_time, 0) << std::endl;
  std::cout << "Total tasks executed:    " << total_tasks << std::endl;

  

  taskman_shutdown();

  return EXIT_SUCCESS;

}
